"""main module for triggering input and output modules"""
# optional system certificate trust
try:
    import truststore
    truststore.inject_into_ssl()
except ImportError:
    pass

# standard modules
try:
    from os import path
    from shutil import move
    import datetime
    import pickle
except ImportError as e:
    print(
        f'Failed to import required module: {e}\n'
        'Do you need to run pip install -r requirements.txt?')
    exit()

# TTSPod modules
from ttspod.remote_sync import sync as rsync
from ttspod.config import Config
from ttspod.content import Content
from ttspod.links import Links
from ttspod.pod import Pod
from ttspod.speech.speech import Speech
from ttspod.ttspocket import TTSPocket
from ttspod.ttsinsta import TTSInsta
from ttspod.wallabag import Wallabag
from ttspod.logger import Logger


class Main(object):
    """main orchestrating object"""

    def __init__(self, debug=False, config_path=None, engine=None,
                 force=False, dry=False, clean=False, logfile=None,
                 quiet=False):
        self.log = Logger(debug=debug, logfile=logfile, quiet=quiet)
        self.config = Config(
            engine=engine, config_path=config_path, log=self.log
        )
        self.p = None
        self.force = force
        self.dry = dry
        self.cache = []
        self.speech = Speech(config=self.config.speech,
                             dry=self.dry, log=self.log)
        self.load_cache(clean=clean)
        self.pod = Pod(self.config.pod, self.p)
        self.pod.config.debug = self.config.debug
        if self.dry:
            self.log.write("dry-run mode")

    def load_cache(self, clean=False):
        """load podcast and cache from pickle if available"""
        if self.config.cache_path:
            try:
                rsync(
                    source=self.config.cache_path,
                    destination=self.config.pickle,
                    debug=self.config.debug,
                    keyfile=self.config.ssh_keyfile,
                    password=self.config.ssh_password,
                    recursive=False
                )
                self.log.write('cache file synced successfully from server')
            except Exception as err:  # pylint: disable=broad-except
                self.log.write(
                    f'something went wrong syncing the cache file {err}', True)
                if "code 23" in str(err):
                    self.log.write(
                        'if this is your first time running TTSPod, '
                        'this is normal since the cache has never been synced',
                        True)
        if clean:
            self.log.write(
                f'moving {self.config.pickle} cache file and starting fresh')
            move(self.config.pickle, self.config.pickle +
                 str(int(datetime.datetime.now().timestamp())))
        if path.exists(self.config.pickle):
            try:
                with open(self.config.pickle, 'rb') as f:
                    [self.cache, self.p] = pickle.load(f)
            except Exception as err:
                raise ValueError(
                    f"failed to open saved data file {f}: {err}") from err
        return True

    def process(self, items):
        """feed items retrieved by input modules to TTS output modules"""
        if not items:
            self.log.write('no items found to process')
            return False
        for item in items[0:self.config.max_articles]:
            (title, content, url) = item
            if url in self.cache and not self.force:
                self.log.write(f'{title} is already in the feed, skipping')
                continue
            if len(content) > self.config.max_length:
                self.log.write(
                    f'{title} is longer than max length of {self.config.max_length}, skipping')
                continue
            self.log.write(f'processing {title}')
            fullpath = self.speech.speechify(title, content)
            if fullpath:
                self.pod.add((url, title, fullpath))
                self.cache.append(url)
            else:
                self.log.write(
                    f'something went wrong processing {title}', True)
        return True

    def save_cache(self):
        """save cache and podcast pickle"""
        try:
            if self.pod:  # only save/sync cache if podcast data exists
                with open(self.config.pickle, 'wb') as f:
                    pickle.dump([self.cache, self.pod.p], f)
                if self.config.cache_path:
                    try:
                        rsync(
                            source=self.config.pickle,
                            destination=self.config.cache_path,
                            keyfile=self.config.ssh_keyfile,
                            debug=self.config.debug,
                            recursive=False,
                            size_only=False
                        )
                        self.log.write(
                            'cache file synced successfully to server')
                    except Exception as err:  # pylint: disable=broad-except
                        self.log.write(
                            f'something went wrong syncing the cache file {err}', True)
            else:
                self.log.write('cache save failed, no podcast data exists')
        except Exception as err:  # pylint: disable=broad-except
            self.log.write(f'cache save failed {err}')

    def process_wallabag(self, tag):
        """process wallabag items matching tag"""
        wallabag = Wallabag(config=self.config.wallabag, log=self.log)
        items = wallabag.get_items(tag)
        return self.process(items)

    def process_link(self, url, title=None):
        """process link content from URL"""
        links = Links(config=self.config.links, log=self.log)
        items = links.get_items(url, title)
        return self.process(items)

    def process_pocket(self, tag='audio'):
        """process pocket items matching tag"""
        links = Links(config=self.config.links, log=self.log)
        p = TTSPocket(config=self.config.pocket, links=links, log=self.log)
        items = p.get_items(tag)
        return self.process(items)

    def process_insta(self, tag):
        """process instapaper items matching tag"""
        links = Links(self.config.links)
        p = TTSInsta(config=self.config.insta, links=links, log=self.log)
        items = p.get_items(tag)
        return self.process(items)

    def process_content(self, text, title=None):
        """process any sort of text content"""
        content = Content(config=self.config.content, log=self.log)
        items = content.get_items(text, title)
        return self.process(items)

    def process_file(self, fname, title=None):
        """process input from files"""
        content = Content(config=self.config.content, log=self.log)
        items = content.process_file(fname, title)
        return self.process(items)

    def finalize(self):
        """finalize session by saving and syncing podcast and cache"""
        if not self.dry:
            self.pod.save()
            self.pod.sync()
            self.save_cache()
        self.log.close()
        return True



"""generate audio using coqui model"""
# optional system certificate trust
try:
    import truststore
    truststore.inject_into_ssl()
except ImportError:
    pass

# standard modules
try:
    from concurrent.futures import ThreadPoolExecutor
    from nltk.tokenize import BlanklineTokenizer
    from pydub import AudioSegment
    from sys import maxsize
    from traceback import format_exc
    import nltk
    import os
    import spacy
    import textwrap
    import warnings
except ImportError as e:
    print(
        f'Failed to import required module: {e}\n'
        'Do you need to run pip install -r requirements.txt?')
    exit()

# optional modules
try:
    from openai import OpenAI
    # necessary for OpenAI TTS streaming
    warnings.filterwarnings("ignore", category=DeprecationWarning)
except ImportError:
    pass
try:
    from elevenlabs.client import ElevenLabs
    from elevenlabs import save
except ImportError:
    pass

# ttspod modules
from ttspod.logger import Logger

MAX_LENGTH = 4096  # hardcoded maximum value for API-based TTS
OPENAI_MODEL = 'tts-1'
OPENAI_VOICE = 'nova'
ELEVEN_MODEL = 'eleven_monolingual_v1'
ELEVEN_VOICE = 'Daniel'
MAX_WORKERS = 10

class Paid:
    """text to OpenAI or Eleven"""

    def __init__(self, config=None, log=None, engine=None,
                 openai_key=None, eleven_key=None,
                 max_workers=None, openai_model=None,
                 openai_voice=None, eleven_model=None,
                 eleven_voice=None, temp_path=None):
        self.log = log if log else Logger(debug=True)
        if isinstance(config,dict):
            self.c = config
        elif config:
            self.c = vars(config)
        else:
            self.c = { }
        self.nlp = self.get_spacy()
        self.engine = engine if engine else self.c.get('engine','')
        self.oai_key = openai_key if openai_key else self.c.get('openai_api_key','')
        self.el_key = eleven_key if eleven_key else self.c.get('eleven_api_key','')
        self.oai_model = openai_model if openai_model else self.c.get('openai_model',OPENAI_MODEL)
        self.oai_voice = openai_voice if openai_voice else self.c.get('openai_voice',OPENAI_VOICE)
        self.el_model = eleven_model if eleven_model else self.c.get('eleven_model',ELEVEN_MODEL)
        self.el_voice = eleven_voice if eleven_voice else self.c.get('eleven_voice',ELEVEN_VOICE)
        self.max_workers = max_workers if max_workers else self.c.get('max_workers',MAX_WORKERS)
        self.temp_path = os.path.join(temp_path if temp_path else self.c.get('temp_path','.'),'')
        match self.engine.lower():
            case "openai":
                self.tts = OpenAI(api_key=self.oai_key)
            case "eleven":
                self.tts = ElevenLabs(api_key=self.el_key)
            case _:
                self.tts = None
                self.log.write(f'no valid paid TTS engine specified: {self.engine}')

    def get_spacy(self):
        """retrieve model for spacy tokenizer"""
        nlp = None
        try:
            if not spacy.util.is_package("en_core_web_lg"):
                self.log.write('downloading spacy language model')
                spacy.cli.download("en_core_web_lg")
            nlp = spacy.load("en_core_web_lg")
            nlp.add_pipe('sentencizer')
            nltk.data.find('tokenizers/punkt_tab')
            self.log.write("nltk found and activated")
            self.nltk = True
        except LookupError:
            try:
                nltk.download('punkt_tab')
                self.log.write("nltk punkt_tab downloaded")
                self.nltk = True
            except Exception:  # pylint: disable=broad-except
                self.log.write("nltk loading failed")
        return nlp

    def segmentize(self, text):
        """ break arbitrary input text of segments of MAX_LENGTH or less"""
        if self.nltk:
            paragraphs = BlanklineTokenizer().tokenize(text)
        else:
            paragraphs = text.split('\n\n')
        segments = []

        for para in paragraphs:
            self.log.write(f"paragraph {para}")
            if len(para) < 8:  # skip very short lines which are likely not text
                continue
            if len(para) > MAX_LENGTH:  # break overlong paras into sentences
                self.log.write(
                    f"further splitting paragraph of length {len(para)}")
                sentences = []
                try:
                    doc = self.nlp(para)
                    sentences = [sent.text.strip() for sent in doc.sents]
                except Exception:  # pylint: disable=broad-except
                    pass
                if not sentences:  # fallback method, simple line wrap
                    sentences = textwrap.wrap(text=para, width=MAX_LENGTH)
                for sentence in sentences:
                    # break sentences greater than 4096 characters into smaller pieces
                    if len(sentence) > MAX_LENGTH:
                        chunks = textwrap.wrap(text=sentence, width=MAX_LENGTH)
                        for chunk in chunks:
                            if len(chunk) < MAX_LENGTH:
                                segments.append(chunk)
                            else:  # if we can't find a small enough piece, we give up
                                self.log.write(
                                    "abnormal sentence fragment found, skipping")
                    else:
                        segments.append(sentence)
            else:
                segments.append(para)
        return segments

    def convert(self, text, output_file):
        """convert text input to given output_file"""

        segments = self.segmentize(text)

        try:
            if self.engine == "openai":
                def tts_function(z):
                    return self.tts.audio.speech.create(
                        model=self.oai_model,
                        voice=self.oai_voice,
                        input=z
                    )
            elif self.engine == "eleven":
                def tts_function(z):
                    return self.tts.generate(
                        voice=self.el_voice,
                        model=self.el_model,
                        text=z
                    )
            else:
                raise ValueError("No TTS engine configured.")
            futures = []
            # TODO - use these hashes to see if any segment has already been transcribed
            self.log.write(f'processing {len(segments)} segments')
            hashes = [str(hash(segment) % ((maxsize + 1) * 2))
                      for segment in segments]
            combined = AudioSegment.empty()
            with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
                for future in executor.map(tts_function, segments):
                    futures.append(future)
                for i, future in enumerate(futures):
                    segment_audio = os.path.join(
                        self.temp_path,
                        f'{output_file}-{hashes[i]}.mp3'
                    )
                    if self.engine == "openai":
                        future.stream_to_file(segment_audio)
                    elif self.engine == "eleven":
                        save(future, segment_audio)
                    combined += AudioSegment.from_mp3(segment_audio)
                combined.export(output_file, format="mp3")
        except Exception as err:  # pylint: disable=broad-except
            self.log.write(
                f'TTS engine {self.engine} failed: {err}\n'+format_exc()
            )
        return True if os.path.isfile(output_file) else False

if __name__ == "__main__":
    paid = Paid()



"""main TTS processor"""
# optional system certificate trust
try:
    import truststore
    truststore.inject_into_ssl()
except ImportError:
    pass

# standard modules
try:
    from anyascii import anyascii
    import os
    import re
    import unicodedata
    import uuid
    import warnings
except ImportError as e:
    print(
        f'Failed to import required module: {e}\n'
        'Do you need to run pip install -r requirements.txt?')
    exit()

# TTSPod modules
from ttspod.logger import Logger
from ttspod.speech.paid import Paid

# optional generator modules
ENGINES = {}
# pylint: disable=unused-import
# pylint: disable=ungrouped-imports
try:
    from openai import OpenAI
    # necessary for OpenAI TTS streaming
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    ENGINES['openai'] = True
except ImportError:
    pass
try:
    from elevenlabs.client import ElevenLabs
    from elevenlabs import save
    ENGINES['eleven'] = True
except ImportError:
    pass
try:
    from ttspod.speech.whisper import Whisper
    ENGINES['whisper'] = True
except ImportError:
    pass
try:
    from ttspod.speech.coqui import Coqui
    ENGINES['coqui'] = True
except ImportError:
    pass
# pylint: enable=unused-import
# pylint: enable=ungrouped-imports


class Speech(object):
    """main TTS processor"""

    def __init__(self, config, dry=False, log=None):
        self.dry = dry
        if dry:
            return
        self.log = log if log else Logger(debug=True)
        self.config = config
        self.config.nltk = False
        self.final_path = config.final_path
        match self.config.engine.lower():
            case "openai" if "openai" in ENGINES:
                self.tts = Paid(config=self.config, log=self.log)
            case "eleven" if "eleven" in ENGINES:
                self.tts = Paid(config=self.config, log=self.log)
            case "whisper" if "whisper" in ENGINES:
                self.tts = Whisper(config=self.config, log=self.log)
            case "coqui" if "coqui" in ENGINES:
                self.tts = Coqui(config=self.config, log=self.log)
            case _:
                raise ValueError('TTS engine not configured')

    def slugify(self, value):
        """convert an arbitrary string to a valid filename"""
        value = str(value)
        value = unicodedata.normalize('NFKD', value).encode(
            'ascii', 'ignore').decode('ascii')
        value = re.sub(r'[^\w\s-]', '', value.lower())
        return re.sub(r'[-\s]+', '-', value).strip('-_')

    def speechify(self, title="missing", raw_text=""):
        """workhorse TTS function"""
        clean_title = self.slugify(title)
        out_file = os.path.join(self.config.final_path, f'{clean_title}.mp3')
        text = anyascii(raw_text)
        temp = str(uuid.uuid4())

        if os.path.exists(out_file):  # don't overwrite existing files
            out_file = os.path.join(
                self.config.final_path, f'{clean_title}-{temp}.mp3')

        if self.dry:  # quit if dry run
            self.log.write(f'dry run: not creating {out_file}')
            return

        self.log.write(self.tts.convert(text=text, output_file=out_file))

        if os.path.exists(out_file):
            os.chmod(out_file, 0o644)
            return out_file



"""generate audio using coqui model"""
# optional system certificate trust
try:
    import truststore
    truststore.inject_into_ssl()
except ImportError:
    pass
try:
    from warnings import simplefilter  # disable coqui future warnings
    simplefilter(action='ignore', category=FutureWarning)
    from contextlib import redirect_stdout, redirect_stderr
    from nltk import sent_tokenize
    from glob import glob
    from os import path, environ as env
    import torch
    import torchaudio
    import re
    from pathlib import Path
    from torch import cuda
    from torch.backends import mps
    from transformers import pytorch_utils
    from whisperspeech.pipeline import Pipeline
    import io
except ImportError as e:
    print(
        f'Failed to import required module: {e}\n'
        'Do you need to run pip install -r requirements.txt?')
    exit()

# ttspod modules
from ttspod.logger import Logger
from ttspod.util import patched_isin_mps_friendly


class Whisper:
    """whisper text to speech generator"""

    def __init__(self, config=None, log=None, t2s_model=None, s2a_model=None, voice=None):
        self.log = log if log else Logger(debug=True)
        self.config = config
        self.voice = None
        if cuda.is_available():
            self.cpu = 'cuda'
        elif mps.is_available():
            self.cpu = 'mps'
            env["PYTORCH_ENABLE_MPS_FALLBACK"] = "1"
            pytorch_utils.isin_mps_friendly = patched_isin_mps_friendly
        else:
            self.cpu = 'cpu'
        if not config:
            c = {}
        else:
            if not isinstance(config, dict):
                c = vars(config)
            else:
                c = config
        t2s_model = c.get(
            'whisper_t2s_model', 'whisperspeech/whisperspeech:t2s-base-en+pl.model')
        s2a_model = c.get(
            'whisper_a2s_model', 'whisperspeech/whisperspeech:s2a-q4-base-en+pl.model')
        voice = path.expanduser(voice if voice else c.get('voice', ''))
        if path.isfile(voice):
            self.voice = voice
        elif path.isdir(voice):
            audio_files = glob(path.join(voice, "*wav"))+glob(path.join(voice, "*mp3"))
            if audio_files:
                self.voice = audio_files[0]
        self.tts = Pipeline(t2s_ref=t2s_model,
                            s2a_ref=s2a_model,
                            device=self.cpu,
                            torch_compile=False,
                            optimize=True)

    def split_and_prepare_text(self, text, cps=14):
        """break text into chunks for whisperspeech"""
        chunks = []
        sentences = sent_tokenize(text)
        chunk = ""
        for sentence in sentences:
            sentence = re.sub('[()]', ",", sentence).strip()
            sentence = re.sub(",+", ",", sentence)
            sentence = re.sub('"+', "", sentence)
            if len(chunk) + len(sentence) < 20*cps:
                chunk += " " + sentence
            elif chunk:
                chunks.append(chunk)
                chunk = sentence
            elif sentence:
                chunks.append(sentence)
        if chunk:
            chunks.append(chunk)
        return chunks

    def whisper_long(self, chunks=None, cps=14, overlap=100, output=None, speaker=None):
        """main whisperspeech generator"""
        if not speaker:
            self.log.write('using default speaker')
            speaker = self.tts.default_speaker
        elif isinstance(speaker, (str, Path)):
            self.log.write(f'extracting speaker {speaker}')
            speaker = self.tts.extract_spk_emb(speaker)
        r = []
        old_stoks = None
        old_atoks = None
        stdout_buffer = io.StringIO()
        stderr_buffer = io.StringIO()
        for i, chunk in enumerate(chunks):
            self.log.write(
                f"processing chunk {i+1} of {len(chunks)}\n"
                "--------------------------\n"
                f"{chunk}\n"
                "--------------------------\n")
            try:
                with redirect_stdout(stdout_buffer), redirect_stderr(stderr_buffer):
                    stoks = self.tts.t2s.generate(
                        chunk, cps=cps, show_progress_bar=False)[0]
                    stoks = stoks[stoks != 512]
                    if old_stoks is not None:
                        assert len(stoks) < 750-overlap  # TODO
                        stoks = torch.cat([old_stoks[-overlap:], stoks])
                        atoks_prompt = old_atoks[:, :, -overlap*3:]
                    else:
                        atoks_prompt = None
                    atoks = self.tts.s2a.generate(
                        stoks,
                        atoks_prompt=atoks_prompt,
                        speakers=speaker.unsqueeze(0),
                        show_progress_bar=False
                    )
                    if atoks_prompt is not None:
                        atoks = atoks[:, :, overlap*3+1:]
                    r.append(atoks)
                    self.tts.vocoder.decode_to_notebook(atoks)
            except Exception as err:  # pylint: disable=broad-except
                self.log.write(f'chunk {i+1} failed with error {err}')
            old_stoks = stoks
            old_atoks = atoks
        audios = []
        for i, atoks in enumerate(r):
            if i != 0:
                audios.append(torch.zeros((1, int(24000*0.5)),
                              dtype=atoks.dtype, device=atoks.device))
            audios.append(self.tts.vocoder.decode(atoks))
        if output:
            torchaudio.save(output, torch.cat(audios, -1).cpu(), 24000)
        return stdout_buffer.getvalue()+"\n"+stderr_buffer.getvalue()

    def convert(self, text, output_file):
        """convert text input to given output_file"""
        chunks = self.split_and_prepare_text(text)
        try:
            results = self.whisper_long(
                chunks=chunks, output=output_file, speaker=self.voice)
            if results:
                self.log.write(
                    f'TTS conversion complete: {results}')
                return True
            else:
                return False
        except Exception as err:  # pylint: disable=broad-except
            self.log.write(f'TTS conversion failed: {err}', True)
            return False


if __name__ == "__main__":
    whisper = Whisper()
